#: default epoch identifier
EPOCH_ID = 'Epoch_idx'

#: default time identifier
TIME = 'Time'

#: default generic channel list
CHANNELS = [
    'ch00',
    'ch01',
    'ch02',
    'ch03',
    'ch04',
    'ch05',
    'ch06',
    'ch07',
    'ch08',
    'ch09',
    'ch10',
    'ch11',
    'ch12',
    'ch13',
    'ch14',
    'ch15',
]
